// users_hwaredriver.c for external frequency control via serial port
// for IC-275/IC-706/FT-1000/FT-736/TS-850/TS-2000
// version: 04 may 2007
// written for linrad 02-31
// runs under windows and linux 
//
// To select a transceiver: uncomment the corresponding define statement
// To generate a windows version in  linux:
//   install mingw32 from Leif's website,
//   save this users_hwaredriver.c as wusers_hwaredriver.c 
//   run ./configure and  make linrad.exe
// To generate a windows version in windows:
//   install mingw.zip and nasm.zip from Leif's website,
//   save this users_hwaredriver.c as wusers_hwaredriver.c 
//   run configure.exe  and  make.bat
//To add your own transceiver:
//   1. Add a statement #define TRANSCEIVER_NAME "MYTRX" ( maximum 7 characters)
//   2. Define the corresponding serial port parameters and frequency range.
//      (Look at line 618 for details )
//   3. Add the corresponding serial protocol
//      (Look at line 720 for details)

//
// Based on work from DL3IAE/IK7EZN/K1JT/ON4IY/ON7EH/W3SZ
// with the support of Leif SM5BSZ
// All errors are mine
// Use at your own risk and enjoy
// Pierre/ON5GN

//#define TRANSCEIVER_NAME "IC-275"
//#define TRANSCEIVER_NAME "IC-706A"    //ICOM-706
//#define TRANSCEIVER_NAME "IC-706B"    //ICOM-706MKII
//#define TRANSCEIVER_NAME "IC-706C"    //ICOM-706MKIIG

//#define TRANSCEIVER_NAME "FT-1000"
#define TRANSCEIVER_NAME "FT-736"

//#define TRANSCEIVER_NAME "TS-850"
//#define TRANSCEIVER_NAME "TS-2000"



// valid ranges: SERPORT_NUMBER = 1 to 4 (1=COM1,4=COM4 )
#define SERPORT_NUMBER 1


// ******************************************************************************
//               ROUTINES FOR A 'MOVABLE' FIXED SIZE USERGRAPH
//             screenposition of usergraph and offset values are 
//                saved in the file par_"TRANSCEIVER_NAME"
//                 The same file is used in all rx modes. 
// ******************************************************************************

#include <string.h>

#define NO_OF_BANDS 12               
float bandlim[]={2.5,  //0 
                 5.0,  //1   
                 9.0,  //2
                12.0,  //3
                16.0,  //4
                19.0,  //5
                23.0,  //6
                26.0,  //7
                40.0,  //8
               100.0,  //9
               300.0,  //10
               BIG};   //11


#define USERS_GRAPH_TYPE1 64
typedef struct {
int ytop;
int ybottom;
int xleft;
int xright;
int offs_hz[NO_OF_BANDS];
int offs_khz[NO_OF_BANDS];
int yborder;    // required for move graph
} UG_PARMS;

#define UG_TOP 0
#define UG_BOTTOM 1
#define UG_LEFT 2
#define UG_RIGHT 3
#define UG_INCREASE_OFFS_HZ 4
#define UG_DECREASE_OFFS_HZ 5
#define UG_INCREASE_OFFS_KHZ 6
#define UG_DECREASE_OFFS_KHZ 7
#define MAX_UGBUTT 8

UG_PARMS ug;
BUTTONS ugbutt[MAX_UGBUTT];

                 

int ug_old_y1;
int ug_old_y2;
int ug_old_x1;
int ug_old_x2;
int users_graph_scro;

int ug_oldx;     // required for move graph
int ug_oldy;     // required for move graph
int ug_yborder;  // required for move graph
int ug_hsiz;     // required for move graph
int ug_vsiz;     // required for move graph 

int ug_band_no;

#define MAX_MSGSIZE 30     //Size of messages. min size=30
char ug_msg0[MAX_MSGSIZE]; //messages in usergraph 
char ug_msg1[MAX_MSGSIZE];
char ug_msg2[MAX_MSGSIZE];
int ug_msg_red;   //switch for message-color in usergraph
char ugfile[20];  //name of usergraph parameter file
double min_transceiver_freq=0;
double max_transceiver_freq=0;



void set_press_q_key_msg(void)
{
ug_msg_red=1;
strcpy(ug_msg0,"PRESS Q-KEY TO SET FREQUENCY"); 
ug_msg1[0]=0;
ug_msg2[0]=0;
}


void check_ug_borders(void)  // required for move graph
{
current_graph_minh=ug_vsiz;
current_graph_minw=ug_hsiz;
check_graph_placement((void*)(&ug));
set_graph_minwidth((void*)(&ug));
}

void show_offset(void)
{
char s[80];
sprintf(s,"Offset-fine(Hz): %d        ",ug.offs_hz[ug_band_no]);
s[25]=0;
lir_pixwrite(ug.xleft+1.5*text_width,ug.ytop+1.5*text_height,s);
sprintf(s,"Offset-raw(Khz): %d        ",ug.offs_khz[ug_band_no]);
s[25]=0;
lir_pixwrite(ug.xleft+1.5*text_width,ug.ytop+1+2.5*text_height,s);
}

void blankfill_ugmsg(char *msg)
{
int i;
i=strlen(msg);
if(i>=MAX_MSGSIZE)lirerr(3413400);
while(i<MAX_MSGSIZE-1)
  {
  msg[i]=' ';
  i++;
  }
msg[MAX_MSGSIZE-1]=0;  
}

void show_user_parms(void)
{
int i;
char s[80];
extern char serport_name[10];
FILE *Fp;
// Show the user parameters on screen
// and issue hardware commands if required.
// Use hware_flag to direct calls to control_hware() when cpu time 
// is available in case whatever you want to do takes too much time 
// to be done immediately.
// Note that mouse control is done in the narrowband processing thread
// and that you can not use lir_sleep to wait for hardware response here.
hide_mouse(ug.xleft, ug.xright,ug.ytop,ug.ybottom);
sprintf(s,"%s %s FREQ. CNTL.",TRANSCEIVER_NAME,serport_name);
settextcolor(14);
lir_pixwrite(ug.xleft+.5*text_width,ug.ytop+.4*text_height,s);
settextcolor(7);
show_offset();

//DISPLAY MESSAGES IN  MESSAGE-AREA OF USERGRAPH

// Fill message with blanks up to character MAX_MSGSIZE-2
blankfill_ugmsg(ug_msg0);
blankfill_ugmsg(ug_msg1);
blankfill_ugmsg(ug_msg2);

if (ug_msg_red==1)
{settextcolor(12);
}
else
{
settextcolor(10);
}
lir_pixwrite(ug.xleft+31*text_width,ug.ytop+.5*text_height,ug_msg0);
lir_pixwrite(ug.xleft+31*text_width,ug.ytop+1.5*text_height,ug_msg1);
lir_pixwrite(ug.xleft+31*text_width,ug.ytop+2.5*text_height,ug_msg2);
settextcolor(7);

//save screenposition and offset-values in usergraph to usergraph parameter file
Fp=fopen(ugfile,"w");
fprintf(Fp,"%d\n%d\n",ug.xleft,ug.ytop);
for(i=0; i<NO_OF_BANDS; i++)fprintf(Fp,"%d\n",ug.offs_hz[i]);
for(i=0; i<NO_OF_BANDS; i++)fprintf(Fp,"%d\n",ug.offs_khz[i]);
fclose(Fp);
}

void users_set_band_no(void)
{
ug_band_no=0;
while(bandlim[ug_band_no] < fg.passband_center)ug_band_no++;
if(all_threads_started == TRUE)
  {
  pause_thread(THREAD_SCREEN);
  set_press_q_key_msg();
  show_user_parms();
  resume_thread(THREAD_SCREEN);
  }
}


void make_users_control_graph(void)
{
pause_thread(THREAD_SCREEN);
// Set a negative number here.
// In case several windows are desired, give them different
// negative numbers and use scro[m].no to decide what to do
// in your mouse_on_users_graph routine.
check_ug_borders();  // required for move graph
scro[users_graph_scro].no=USERS_GRAPH_TYPE1;
// These are the coordinates of the border lines.
scro[users_graph_scro].x1=ug.xleft;
scro[users_graph_scro].x2=ug.xright;
scro[users_graph_scro].y1=ug.ytop;
scro[users_graph_scro].y2=ug.ybottom;
// Each border line is treated as a button.
// That is for the mouse to get hold of them so the window can be moved.
ugbutt[UG_LEFT].x1=ug.xleft;
ugbutt[UG_LEFT].x2=ug.xleft;
ugbutt[UG_LEFT].y1=ug.ytop;
ugbutt[UG_LEFT].y2=ug.ybottom;
ugbutt[UG_RIGHT].x1=ug.xright;
ugbutt[UG_RIGHT].x2=ug.xright;
ugbutt[UG_RIGHT].y1=ug.ytop;
ugbutt[UG_RIGHT].y2=ug.ybottom;
ugbutt[UG_TOP].x1=ug.xleft;
ugbutt[UG_TOP].x2=ug.xright;
ugbutt[UG_TOP].y1=ug.ytop;
ugbutt[UG_TOP].y2=ug.ytop;
ugbutt[UG_BOTTOM].x1=ug.xleft;
ugbutt[UG_BOTTOM].x2=ug.xright;
ugbutt[UG_BOTTOM].y1=ug.ybottom;
ugbutt[UG_BOTTOM].y2=ug.ybottom;
// Draw the border lines
graph_borders((void*)&ug,7);
ug_oldx=-10000;                   //from freq_control
settextcolor(7);
make_button(ug.xleft+27.5*text_width+2,ug.ybottom-3*text_height/2-1,
                                         ugbutt,UG_DECREASE_OFFS_HZ,25);
make_button(ug.xleft+29*text_width+2,ug.ybottom-3*text_height/2-1,
                                     ugbutt,UG_INCREASE_OFFS_HZ,24); 

make_button(ug.xleft+27.5*text_width+2,ug.ybottom-1*text_height/2-1,
                                         ugbutt,UG_DECREASE_OFFS_KHZ,25);
make_button(ug.xleft+29*text_width+2,ug.ybottom-1*text_height/2-1,
                                     ugbutt,UG_INCREASE_OFFS_KHZ,24); 
show_user_parms();
//draw separatorlines in usergraph
lir_line(ug.xleft+30.5*text_width,ug.ytop,ug.xleft+30.5*text_width,ug.ytop+3.5*text_height,7);
lir_line(ug.xleft,ug.ytop+1.3*text_height,ug.xleft+30.5*text_width,ug.ytop+1.3*text_height,7);
//
resume_thread(THREAD_SCREEN);
}


void mouse_continue_users_graph(void)
{
int i, j;      // required for move graph

switch (mouse_active_flag-1)
  {
// Move border lines immediately.
// for other functions, wait until button is released.
// Move fixed size window  based on coding in freq_control.c
case UG_TOP:
  if(ug.ytop!=mouse_y)goto ugm;
  break;

  case UG_BOTTOM:
  if(ug.ybottom!=mouse_y)goto ugm;
  break;

  case UG_LEFT:
  if(ug.xleft!=mouse_x)goto ugm;
  break;

  case UG_RIGHT:
  if(ug.xright==mouse_x)break;
ugm:;
  pause_screen_and_hide_mouse();
  graph_borders((void*)&ug,0);
  if(ug_oldx==-10000)
    {
    ug_oldx=mouse_x;
    ug_oldy=mouse_y;
    }
  else
    {
    i=mouse_x-ug_oldx;
    j=mouse_y-ug_oldy;  
    ug_oldx=mouse_x;
    ug_oldy=mouse_y;
    ug.ytop+=j;
    ug.ybottom+=j;
    ug.xleft+=i;
    ug.xright+=i;
    check_ug_borders();    // check borders
    ug.yborder=(ug.ytop+ug.ybottom)>>1;
    }
  graph_borders((void*)&ug,15); 
  resume_thread(THREAD_SCREEN);
  break;

  default:
  goto await_release;
  }
if(leftpressed == BUTTON_RELEASED)goto finish;
return;
await_release:;
if(leftpressed != BUTTON_RELEASED) return;
// Assuming the user wants to control hardware we
// allow commands only when data is not over the network.
if((ui.network_flag&NET_RX_INPUT) == 0) //for linrad02.23
  {
  switch (mouse_active_flag-1)
    {
    case UG_INCREASE_OFFS_HZ: 
    ug.offs_hz[ug_band_no]++;
    break;

    case UG_DECREASE_OFFS_HZ:
    ug.offs_hz[ug_band_no]--;
    break;

    case UG_INCREASE_OFFS_KHZ: 
    ug.offs_khz[ug_band_no]++;
    break;

    case UG_DECREASE_OFFS_KHZ:
    ug.offs_khz[ug_band_no]--;
    break;

    default:
// This should never happen.    
    lirerr(211053);
    break;
    }
  set_press_q_key_msg();
  }  
finish:;
hide_mouse(ug_old_x1,ug_old_x2,ug_old_y1,ug_old_y2);
leftpressed=BUTTON_IDLE;  
mouse_active_flag=0;
graph_borders((void*)&ug,0);
lir_fillbox(ug_old_x1,ug_old_y1,ug_old_x2-ug_old_x1,ug_old_y2-ug_old_y1,0);
make_users_control_graph();
ug_oldx=-10000; 
}

void new_user_offs_hz(void)
{
ug.offs_hz[ug_band_no]=numinput_int_data;
pause_thread(THREAD_SCREEN);
set_press_q_key_msg();
show_user_parms();
resume_thread(THREAD_SCREEN);
}

void new_user_offs_khz(void)
{
ug.offs_khz[ug_band_no]=numinput_float_data;
pause_thread(THREAD_SCREEN);
set_press_q_key_msg();
show_user_parms();
resume_thread(THREAD_SCREEN);
}


void mouse_on_users_graph(void)
{
int event_no;
// First find out if we are on a button or border line.
for(event_no=0; event_no<MAX_UGBUTT; event_no++)
  {
  if( ugbutt[event_no].x1 <= mouse_x && 
      ugbutt[event_no].x2 >= mouse_x &&      
      ugbutt[event_no].y1 <= mouse_y && 
      ugbutt[event_no].y2 >= mouse_y) 
    {
    ug_old_y1=ug.ytop;
    ug_old_y2=ug.ybottom;
    ug_old_x1=ug.xleft;
    ug_old_x2=ug.xright;
    mouse_active_flag=1+event_no;
    current_mouse_activity=mouse_continue_users_graph;
    return;
    }
  }
// Not button or border.
// Prompt the user for offs_hz or offs_khz depending on whether the mouse is
// in the upper or the lower part of the window.
mouse_active_flag=1;
numinput_xpix=ug.xleft+18.5*text_width;
if(mouse_x > ug.xleft+18.5*text_width && mouse_x < ug.xleft+25*text_width)
  {
  if(mouse_y > (ug.ytop+2*text_height))
    {
    numinput_ypix=ug.ytop+(2.5*text_height)+1;
    numinput_chars=7;
    erase_numinput_txt();
    numinput_flag=FIXED_FLOAT_PARM;
    par_from_keyboard_routine=new_user_offs_khz; //offset-raw
    }
  else
    {
    numinput_ypix=ug.ytop+1.5*text_height;
    numinput_chars=4;
    erase_numinput_txt();
    numinput_flag=FIXED_INT_PARM;
    par_from_keyboard_routine=new_user_offs_hz; //offset-fine
    }  
  }
else
  {
// If we did not select a numeric input by setting numinput_flag
// we have to set mouse_active flag.
// Set the routine to mouse_nothing, we just want to
// set flags when the mouse button is released.    
  current_mouse_activity=mouse_nothing;
  mouse_active_flag=1;
  }
}

void init_users_control_window(void)
{
int i;
// Set graph position and offset values by  reading usergraph  file
// 
FILE *Fp;
sprintf(ugfile,"%s_ug",rxpar_filenames[rx_mode]);
Fp=fopen(ugfile,"r");
if (Fp==NULL)          // if file not found, use default-values 
  {
  ug.xleft=60*text_width;       //default-value for initial x position 
  ug.ytop=(screen_last_line-4)*text_height;  //default-value for initial y position
  for(i=0; i<NO_OF_BANDS; i++)
    {
    ug.offs_hz[i]=0;       //default-value for initial offset-fine
    ug.offs_khz[i]=0;      //default-value for initial offset-raw
    }
  }  
else
  {
  fscanf(Fp,"%d%d",&ug.xleft,&ug.ytop);
  for(i=0; i<NO_OF_BANDS; i++)fscanf(Fp,"%d",&ug.offs_hz[i]);
  for(i=0; i<NO_OF_BANDS; i++)fscanf(Fp,"%d",&ug.offs_khz[i]);
  fclose(Fp);
  }
set_press_q_key_msg();
ug_hsiz=(30+MAX_MSGSIZE)*text_width;      // required for move graph
ug_vsiz=(3.5*text_height);                // required for move graph
ug.xright=ug.xleft+ug_hsiz;
ug.ybottom= ug.ytop+ug_vsiz;
if(rx_mode < MODE_TXTEST)
  {
  users_graph_scro=no_of_scro;
  make_users_control_graph();
  no_of_scro++;
  if(no_of_scro >= MAX_SCRO)lirerr(89);
  }
}



void users_init_mode(void)
{
// A switch statement can be used to do different things depending on
// the processing mode.
//switch (rx_mode)
//  {
//  case MODE_WCW:
//  case MODE_HSMS:
//  case MODE_QRSS:
//  break;
//  case MODE_NCW:
//  case MODE_SSB:
// Open a window to allow mouse control of your own things
  init_users_control_window();  
//  break;
//  case MODE_FM:
//  case MODE_AM:
//  case MODE_TXTEST:
//  case MODE_RX_ADTEST:
//  case MODE_TUNE:
//  break;
//  }
}
// *************************************************************
//          END ROUTINES FOR 'MOVABLE' FIXED SIZE USERGRAPH
// *************************************************************

void users_eme(void)
{
}

void userdefined_u(void)
{
}



void edit_ugmsg(char *s, double fq)
{
double t1;
int i, j, k, m;
// We are called with a frequency in Hz and already something written
// in the msg string. Find out how many characters we have at our disposal.
j=strlen(s);
i=MAX_MSGSIZE-j-3;
// We will insert separators for each group of 3 digits so we loose i/3 
// positions.
i-=i/3;
// Find out the largest fq we can write in Hz. 
// Write in kHz if fq is above the limit.
t1=i;
t1=pow(10.0,t1)-1;
if(fabs(fq) < t1)
  {
  sprintf (&s[j],"%.0f Hz",fq);
  }
else
  {
  sprintf(&s[j],"%.0f kHz",fq/1000);
  }
// Locate the blank in front of Hz or kHz.  
i=strlen(s);
k=i;
while(s[i] != ' ')i--;
// Insert a separator for every 3 digits.
i--;
j=0;
while(s[i-1] != ' ' && s[i-1] != '-')
  {
  j++;
  if(j==3)
    {
    k++;
    m=k;
    while(m > i)
      {
      s[m]=s[m-1];
      m--;
      }
    s[i]=39;
    j=0;
    }
  i--;
  }
// Align frequency information  to the right
j=MAX_MSGSIZE-1;
s[j]='\x0';
k=j;
while(s[k] != 'z')k--;
if (k!=j-1){
m=j;
while(s[m] != '=')m--;
j--;
for (i=k; i > m; i--) {
   s[j] = s[i];
   s[i] = 32;
   j--;
   }
for (i=j; i > m; i--) {
   s[i] = 32;
   j--;
   }
 }
}


void users_close_devices(void)
{
lir_close_serport();
}

void users_open_devices(void)
{
int serport_baudrate;
int serport_stopbits=1;
// The parallel port is for the WSE units.
if(ui.parport == 0)
  {
  allow_parport=FALSE;
  }
else
  {
  if(lir_parport_permission()==TRUE)
    {
    allow_parport=TRUE;
    }
  else  
    {
    allow_parport=FALSE;
    }
  }  
// initialize serial port parameters and frequency-range 
// according TRANSCEIVER_NAME
//               serport_baudrate = 110 to 57600 ( look at lxsys.c )
//               serport_stopbits = 0 or 1
//               min and max transceiver_frequency(Hz) : check your manual
if (TRANSCEIVER_NAME=="IC-275")
  {
  serport_baudrate=9600;
  serport_stopbits=0;
  min_transceiver_freq=144000000;
  max_transceiver_freq=146000000;
  }
if (  (TRANSCEIVER_NAME=="IC-706A")
     |(TRANSCEIVER_NAME=="IC-706B")
     |(TRANSCEIVER_NAME=="IC-706C") )
  {
  serport_baudrate=9600;
  serport_stopbits=0;
  min_transceiver_freq=1800000;
  max_transceiver_freq=148000000;
  }
if (TRANSCEIVER_NAME=="TS-850")
  {
  serport_baudrate=4800;
  serport_stopbits=1;
  min_transceiver_freq=1600000;
  max_transceiver_freq=30000000;
  }
if (TRANSCEIVER_NAME=="TS-2000")
  {
  serport_baudrate=9600;
  serport_stopbits=0;
  min_transceiver_freq=1600000;
  max_transceiver_freq=148000000;
  }
if (TRANSCEIVER_NAME=="FT-736")
  {
  serport_baudrate=4800;
  serport_stopbits=1;
  min_transceiver_freq=144000000;
  max_transceiver_freq=146000000;
  }
if (TRANSCEIVER_NAME=="FT-1000")
  {
  serport_baudrate=4800;
  serport_stopbits=1;
  min_transceiver_freq=1600000;
  max_transceiver_freq=30000000;
  }
lir_open_serport(SERPORT_NUMBER,        // 1=COM1 or /dev/ttyS0 , 2=COM2....
                 serport_baudrate,      // baudrate 110 to 57600
                 serport_stopbits);     // 1=two stopbits, 0=one stopbit
}



void userdefined_q(void)
{
// **************************************************************************
//    ROUTINES FOR EXTERNAL FREQUENCY CONTROL VIA SERIAL PORT 
//     These routines are  called when the 'Q' key is pressed
// **************************************************************************
double transceiver_freq;
double offset_hwfreq;
unsigned long freq;
double transverter_offset_raw;  //raw offset in Khz (for tranverter) 
double transverter_offset_fine; //fine offset in Hz (for transverter)
extern char serport_name[10];
unsigned char a;
int n=-1;
int i;
char s[80];

//INITIALIZE

//initialize message-area in usergraph
ug_msg_red=1;
strcpy(ug_msg0,"FREQUENCY SETTING IN PROGRESS");
ug_msg1[0]=0;
ug_msg2[0]=0;
show_user_parms();


//EXIT IF NO LINRAD-FREQUENCY IS SELECTED
 
if(mix1_selfreq[0] < 0)  
{
  ug_msg_red=1;
  sprintf (ug_msg0,"NO FREQ. SELECT. ON W/F GRAPH");
  goto userdefined_q_exit_1;
}

// EXIT IF INVALID  TRANSCEIVER-FREQUENCY 

transverter_offset_raw=ug.offs_khz[ug_band_no];    //raw offset in Khz for transverter 
transverter_offset_fine=ug.offs_hz[ug_band_no];    //fine offset in Hz for transverter 
offset_hwfreq= transverter_offset_raw +(transverter_offset_fine*.001) ; //Freq in KHz 
transceiver_freq= (hwfreq+offset_hwfreq)*1000;
transceiver_freq=floor(transceiver_freq);                               //Freq in Hz 
freq= transceiver_freq;

if ((transceiver_freq < min_transceiver_freq )
     ||(transceiver_freq > max_transceiver_freq))
{
  ug_msg_red=1;
  sprintf (ug_msg0,"INVALID %s FREQ.",TRANSCEIVER_NAME); 
  goto userdefined_q_exit_2;
}


//OPEN SERIAL PORT 

ug_msg_red=0;
sprintf (ug_msg0,"OPEN %s  OK",serport_name );


// ISSUE  'SET FREQUENCY COMMAND' ACCORDING  TRANSCEIVER_NAME
//    input: freq or transceiver_freq
//    proces:convert input to 'set frequency command' message string 
//    output: use 'n=lir_write_serport(message string,message_length);'



// ***********************IC-275/IC-706 ROUTINE******************** 

if ((TRANSCEIVER_NAME=="IC-275")
    |(TRANSCEIVER_NAME=="IC-706A")
    |(TRANSCEIVER_NAME=="IC-706B")
    |(TRANSCEIVER_NAME=="IC-706C")
    )
{

struct {            // 'Set Frequency message' for IC-275/IC-706
   char prefix[2];
   char radio_addr;
   char computer_addr;
   char cmd_nr;
   unsigned char freq[5];
   char suffix;
   } msg0 = {0xFE, 0xFE,               // preamble code x 2
             0x10,                     // receive address for IC-275
             0xE0,                     // transmit address of controller
             0x05,                     // 'set freq' cmnd code
             0x00,0x00,0x00,0x00,0x00, // freq
             // 145.678.900 Hz  is encoded as x00 x89 x67 x45 x01
             0xFD};                    // end of message code

if (TRANSCEIVER_NAME=="IC-706A")
   msg0.radio_addr='\x48';   //  receive address for IC-706

if (TRANSCEIVER_NAME=="IC-706B")
   msg0.radio_addr='\x4e';   // receive address for IC-706MKII

if (TRANSCEIVER_NAME=="IC-706C")
   msg0.radio_addr='\x58';   // receive address for IC-706MKIIG

//convert 10 digits of freq to packed bcd and store in msg0.freq,
//put first digit-pair in FIRST output byte
for (i=0; i < 5; i++) {
   a = freq%10;
   freq /= 10;
   a |= (freq%10)<<4;
   freq /= 10;
   msg0.freq[i]= a;
//printf("\n %#4.2x ",msg0.freq[i]); //trace output to screen
   }
//printf("\n");      //'next line' to display last trace entry
// write 'Set Frequency message' to serial port
n=lir_write_serport((char *) &msg0, sizeof(msg0));
}

// ***********************TS850/TS-2000 ROUTINE****************


if ((TRANSCEIVER_NAME=="TS-850")
    |(TRANSCEIVER_NAME=="TS-2000"))

{
// "VFO-A frequency" command 
// Freq in Hz; 11 digits, leading zeros, no decimal point, trailing semicolon.
sprintf (s,"FA%011.0f;",transceiver_freq);
n=lir_write_serport(s,14);
//printf("\n %s",s); //trace output to screen
//printf("\n");      //'next line' to display last trace entry
}

// ************************FT736 ROUTINE**********************

if (TRANSCEIVER_NAME=="FT-736")
{
unsigned char open_cmnd[5] =  { 0x00, 0x00, 0x00, 0x00, 0x00};
unsigned char close_cmnd[5] = { 0x80, 0x80, 0x80, 0x80, 0x80};
unsigned char setfreq_cmnd[5]= { 0x00, 0x00, 0x00, 0x00, 0x01};
// 145.678.900 HZ is encoded as x14 x56 x78 x90, last byte 0x01 is set freq cmnd code 
unsigned char cat_cmnd[5];

int write_cat_cmnd(void)
{
int r;
for (i=0;i<5;i=i+1)
 {
  r=lir_write_serport((char*)&cat_cmnd[i],1);
  if (r<0) return -1;
// printf("\n %#4.2x ",cat_cmnd[i]);  //trace output to screen
  lir_sleep(50000); // wait 50msec between each byte
 }
return 0;
}

//open CAT system
memcpy(cat_cmnd,open_cmnd, 5);
n=write_cat_cmnd();
if (n<0) goto ft736_exit;

//discard least significant digit of freq
//convert next 8 digits of freq to packed bcd and store in setfreq_cmd
//put first digit-pair in LAST  output byte 
freq /= 10;
for (i=0; i < 4; i++) {
   a = freq%10;
   freq /= 10;
   a |= (freq%10)<<4;
   freq /= 10;
   setfreq_cmnd[3-i]= a;
   }

//write setfreq_cmd to the serial port
memcpy(cat_cmnd,setfreq_cmnd, 5);
n=write_cat_cmnd();
if (n<0) goto ft736_exit;

//close CAT system
memcpy(cat_cmnd,close_cmnd, 5);
n=write_cat_cmnd();
ft736_exit:;
//printf("\n");  //'next line' to display last trace entry
}

// ***************************FT-1000 ROUTINE*********************

if (TRANSCEIVER_NAME=="FT-1000")
{
unsigned char setfreq_cmnd[5]= { 0x00, 0x00, 0x00, 0x00, 0x0a};
// 014.250.000 Hz is encoded as x00 x50 x42 x01,last byte 0x0a is set freq cmnd code

//discard least significant digit of freq
//convert next 8 digits of freq to packed bcd and store in setfreq_cmd,
//put first digit-pair in FIRST output byte 
freq /= 10;
for (i=0; i < 4; i++) {
   a = freq%10;
   freq /= 10;
   a |= (freq%10)<<4;
   freq /= 10;
   setfreq_cmnd[i]= a;
   }
 
//write setfreq_cmnd to the serial port
for (i=0;i<5;i=i+1)
 {
  n=lir_write_serport((char*)&setfreq_cmnd[i],1);
  if (n<0)  goto ft1000_exit;
//printf("\n %#4.2x ",setfreq_cmnd[i]);  //trace output to screen
  lir_sleep(5000); // wait 5msec between each write
 }
ft1000_exit:;
//printf("\n");  //'next line' to display last trace entry
}

// *************END 'SET FREQUENCY COMMAND' ROUTINES ********

//DISPLAY MESSAGES IN USERGRAPH 

if (n<0)
  {
  ug_msg_red=1;
  sprintf (ug_msg0,"WRITE TO %s FAILED",serport_name);
  goto userdefined_q_exit_1;
  }

ug_msg_red=0;
sprintf (ug_msg0,"LINRAD FREQ = ");
i=strlen(ug_msg0);
edit_ugmsg(ug_msg0,floor(hwfreq*1000));
userdefined_q_exit_2:;
sprintf (ug_msg1,"TOTAL OFFSET= ");
i=strlen(ug_msg1);
edit_ugmsg(ug_msg1,offset_hwfreq*1000);
sprintf (ug_msg2,"%s FREQ= ",TRANSCEIVER_NAME);
edit_ugmsg(ug_msg2, transceiver_freq);

userdefined_q_exit_1:;
show_user_parms();
}
// **************************************************************************
// **** END ROUTINES FOR FREQUENCY SETTING OF TRANSCEIVER VIA SERIAL PORT ***
// **************************************************************************

// **********************************************************
//         WSE Rx and Tx and SDR-14/SDR-IQ Rx routines
// **********************************************************

#include "wse_sdrxx.c"

